/**
 * @file main.c
 * @brief Este projeto demonstra o uso de funcoes de conversao entre tipos de dados envolvendo
 * strings, exibindo os valores convertidos no LCD
 * @author Wu Shin Ting
 * @date 19/10/2020
 */

#include "stdlib.h"
#include "string.h"
#include "math.h"
#include "derivative.h" /* include peripheral declarations */

#define GPIO_PIN(x) (1<<x)

//==================================
// Funcoes utilitarias
//==================================
/**
 * @brief atraso em numero de interacoes <i>
 * @param[in] i numero de iteracoes
 */
void delay (uint32_t i)
{
	__asm("outerLoop:;"
				"sub %0, %0, #1;"
				"mov r2, #11;"
			"innerLoop:;"
				"sub r2, r2, #1;"
				"cmp r2, #0;"
				"bne innerLoop;"
				"cmp %0, #0;"
				"bne outerLoop;"
			: 
			: "r" (i)
			);
}

/**
 * @brief convert a unsigned integer number into a string
 * @param[in] n number
 * @param[in] b system base
 * @param[out] string converted number
 */
char *itoa (int n, char *string, char b)
{
	char sinal;
	unsigned short i, j, a;
	unsigned int m;
	
	if (n == 0) {
		//Tratamento do caso especial
		string[0] = '0';
		string[1] = '\0';
		return string;
	}
	
	if (n < 0) sinal = 1;
	else sinal = 0;
	
	// Converter para codigos ASCII, digito por digito, de menos 
	// significativo para o mais
	i = 0;
	m = abs(n);
	
	while (m != 0) {
		a = m%b;
		string[i+sinal] = (a>9)?(a-10)+'A' : a+'0';
		i++;
		m = m/b;
	}
	string[i+sinal] = '\0';			// terminador
	
	//Reverter a sequencia
	j = strlen(string)-1;
	i = sinal;
			
	while (j > i) {
		char tmp = string[i];
		string[i] = string[j];
		string[j] = tmp;
		i++;
		j--;
	}
	
	// Adicionar o sinal negativo caso necessario
	if (sinal) string[0] = '-';
	
	return string;
}

/**
 * @brief convert a unsigned integer number into a string
 * @param[in] n number
 * @param[in] b system base
 * @param[out] string converted number
 */
char *ftoa (float v, char *string)
{
	unsigned short i;
	unsigned int m, pot;
	int parteI;
	float frac, mfrac;
	
	if (v == 0) {
		//Tratamento do caso especial
		string[0] = '0';
		string[0] = '.';
		string[1] = '\0';
		return string;
	}
	
	// Extrair a parte inteira
	parteI = v/1;
	
	// Extrair a parte fracionaria
	frac = fabs(v - parteI);

	// Converter a parte inteira 
	itoa (parteI, string, 10);	
	
	i = strlen(string);
	
	string[i++] = '.';
	
	// Converter a parte fracionaria para inteira
	pot = 10;
	m = (unsigned int)(frac*pot);
	while (m==0) {
		string[i++] = '0';
		pot *= 10;
		m = (unsigned int)(frac*pot);		
	}
	mfrac = frac*pot;
	while ((mfrac-m) > 0.001) {
		pot *= 10;
		m = (unsigned int)(frac*pot);
		mfrac = frac*pot;		
	}

	// Converter a parte fracionaria 
	itoa (m, &string[i], 10);	

	// Recomputar o tamanho da string
	i = strlen(string);
	
	string[i] = '\0';
	
	return string;
}

//==================================
// LCD
//==================================
/**
 * @brief Funcao do byte
 */
typedef enum lcd_RS_tag {
	COMANDO,  /**< Comando */
	DADO	  /**< Dado */
} lcd_RS_type;

/**
 * @brief Inicializa os pinos conectados ao latch 74573 onde estao conectados os 8 leds vermelhos
 */
void GPIO_initLCD(){
	SIM_SCGC5 |= SIM_SCGC5_PORTC_MASK ;     // Habilita o clock da PORTA C

	// Configura os pinos PORTC[10:0] como GPIO de saida 
	PORTC_PCR0 |= PORT_PCR_MUX(0x1);       // D0-D7 dos dados (GPIO)
	PORTC_PCR1 |= PORT_PCR_MUX(0x1);
	PORTC_PCR2 |= PORT_PCR_MUX(0x1);
	PORTC_PCR3 |= PORT_PCR_MUX(0x1);
	PORTC_PCR4 |= PORT_PCR_MUX(0x1);
	PORTC_PCR5 |= PORT_PCR_MUX(0x1);
	PORTC_PCR6 |= PORT_PCR_MUX(0x1);
	PORTC_PCR7 |= PORT_PCR_MUX(0x1);
	
	PORTC_PCR8 |= PORT_PCR_MUX(0x1);       // RS do LCD
	PORTC_PCR9 |= PORT_PCR_MUX(0x1);       // E do LCD

	GPIOC_PDDR |= GPIO_PDDR_PDD(GPIO_PIN(9)|GPIO_PIN(8) |GPIO_PIN(7) | 
			GPIO_PIN(6) | GPIO_PIN(5) | GPIO_PIN(4) | GPIO_PIN(3) | 
			GPIO_PIN(2) | GPIO_PIN(1) | GPIO_PIN(0));       // Configura como saidas
	GPIOC_PDOR &= ~GPIO_PDOR_PDO(GPIO_PIN(9) |GPIO_PIN(8)|GPIO_PIN(7) | 
			GPIO_PIN(6) | GPIO_PIN(5) | GPIO_PIN(4) | GPIO_PIN(3) | 
			GPIO_PIN(2) | GPIO_PIN(1) | GPIO_PIN(0)); // Inicializa os pinos com 0
}

/**
 * @brief Escrever nos pinos de dados um caractere de 8 bits
 * @param[in] c caractere
 */
void GPIO_setByte(char c) {
	// colocar os dados nos pinos de dados DO-D7
	GPIOC_PDOR &= 0xFFFFFF00; 		// zerar o byte menos significativo
	GPIOC_PDOR |= (unsigned int) c; // setas os bits em 1 no byte menos significativo	
}

/**
 * @brief Gerar o pulso de Enable de ~450ns
 */
void GPIO_enableE()
{
	// gerar um pulso  (~ 450 ns)
	GPIOC_PSOR = GPIO_PIN (9);		// setar em 1
	delay(1);
	GPIOC_PCOR = GPIO_PIN (9);		// resetar em 0
}

/**
 * @brief Setar o tipo de transferencia
 * @param[in] i tipo
 */
void GPIO_setRS (lcd_RS_type i) {
	if(i == COMANDO) 
		GPIOC_PCOR = GPIO_PIN(8);     // Seta o LCD no modo de comando		
	else if (i == DADO) 
		GPIOC_PSOR = GPIO_PIN(8);    // Seta o LCD no modo de dados
}

/**
 * @brief Transferir um byte ao LCD
 * @param[in] c byte
 * @param[in] t tempo de processamento em multiplos de 2 us
 */
void escreveLCD(char c, uint32_t t)
{
	GPIO_setByte(c);
	GPIO_enableE();
	delay(t);
}

/**
 * @brief Inicializar o LCD
 */
void initLCD () {
	//espera por mais de 30ms = 30000us/2 = 15000
	delay(15000);
	
	GPIO_setRS(COMANDO);
	escreveLCD(0x38, 20);			//Function Set: 39us/2 -> 20
	escreveLCD(0x0C, 20);		    //Display ON/OFF Control: 39us/2 -> 20
	escreveLCD(0x01, 765);			//Display Clear: 1530us/2 -> 765
	escreveLCD(0x06, 20);			//Entry mode set: 39us/2 -> 20
}

/**
 * @brief Limpe o display
 */
void limpeDisplayLCD () {
	GPIO_setRS(COMANDO);
	escreveLCD(0x01, 765);			//Display Clear: 1530us/2 -> 765
}

/**
 * @brief Escrever uma mensagem str no LCD a partir do endereco especificado
 * @param[in] end endereco DDRAM
 * @param[in] str mensagem
 */
void escreveMensagem (uint8_t end, char* str)
{
	//Posicione o cursor no end
	uint8_t tmp = 0b10000000 | end; // seta o bit 7 em 1
	GPIO_setRS(COMANDO);			// comando
	escreveLCD(tmp, 20);			// seta o endereco do cursor

	//Chaveia para modo DADO
	GPIO_setRS (DADO);
	while (*str) {
		escreveLCD(*str, 20);
		str++;
	}
}

/**
 * @brief Escrever um bitmap na memoria CGRAM a partir do endereco especificado
 * @param[in] end endereco CGRAM
 * @param[in] str bitmap
 */
void escreveBitmap (uint8_t end, char* bitmap)
{
	//Posicione o cursor no end
	uint8_t tmp = 0b01000000 | (end*8); // seta o bit 7 em 0 e bit 6 em 1
	GPIO_setRS(COMANDO);				// comando
	escreveLCD(tmp, 20);				// seta o endereco do cursor

	//Chaveia para modo DADO
	GPIO_setRS (DADO);
	uint8_t i;
	for (i=0; i< 8; i++) {
		escreveLCD(bitmap[i], 20);
	}
}

/**
 * @brief Fluxo de controle principal
 */
int main(void)
{
	int a, b;
	float fa, fb, fc;
	
   	char result[50]; 
	   	
   	GPIO_initLCD();
   	initLCD();
   	
   	itoa (47856, result, 16);
   	escreveMensagem (0x00, result);   	
   	a = atoi (result);
   	
   	itoa (3856, result, 10);
   	escreveMensagem (0x40, result);
   	b = atoi (result);
   	
   	itoa (-2790, result, 10);
   	escreveMensagem (0x40, result);
   	b = atoi (result);

   	ftoa (4.578, result);
   	escreveMensagem (0x40, result);
  	fa = atof(result);

   	ftoa (-65.32914, result);
   	escreveMensagem (0x40, result);
  	fb = atof(result);

   	ftoa (0.0060321, result);
   	escreveMensagem (0x40, result);
  	fc = atof(result);

   	for(;;) {	   

   	}
   	
	return 0;
}
